# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import time

import click
from tqdm import tqdm

from fastwarc import FastWARCError
from fastwarc.benchmark import benchmark
from fastwarc.tools import CompressionAlg
import fastwarc.tools as tools


def exception_handler(exctype, value, _):
    if exctype == FastWARCError:
        click.echo(str(value), err=True)
        sys.exit(1)

    raise value


sys.excepthook = exception_handler


@click.group(context_settings=dict(help_option_names=['-h', '--help']))
def main():
    """FastWARC Command Line Interface."""
    return 0


def _human_readable_bytes(byte_num):
    for unit in ['bytes', 'KiB', 'MiB', 'GiB', 'TiB']:
        if byte_num <= 1024 or unit == 'TiB':
            if int(byte_num) == byte_num:
                return f'{byte_num:d} {unit}'
            return f'{byte_num:.2f} {unit}'
        byte_num /= 1024


@main.command()
@click.argument('infile', type=click.Path(dir_okay=False, exists=True))
@click.argument('outfile', type=click.Path(dir_okay=False, exists=False))
@click.option('-c', '--compress-alg', type=click.Choice(['gzip', 'lz4', 'uncompressed', 'auto']),
              default='auto', show_default=True, help='Compression algorithm to use for output file')
@click.option('-d', '--decompress-alg', type=click.Choice(['gzip', 'lz4', 'uncompressed', 'auto']),
              default='auto', show_default=True,
              help='Decompression algorithm for decoding input file (auto tries to detect based on file extension)')
@click.option('-l', '--compress-level', type=int, default=None, help='Compression level (defaults to max)')
@click.option('-q', '--quiet', is_flag=True, help='Do not print progress information')
def recompress(infile, outfile, compress_alg, decompress_alg, compress_level, quiet):
    """Recompress a WARC file with different settings."""

    compress_alg = getattr(CompressionAlg, compress_alg)
    decompress_alg = getattr(CompressionAlg, decompress_alg)

    comp_args = {}
    if compress_level is None:
        if compress_alg == CompressionAlg.gzip:
            comp_args['compression_level'] = 9
        elif compress_alg == CompressionAlg.lz4:
            comp_args['compression_level'] = 12

    bytes_written = 0
    num = 0
    start = time.time()
    try:
        for _, b in tqdm(tools.recompress_warc_interactive(infile, outfile, decompress_alg, compress_alg, **comp_args),
                         desc='Recompressing WARC file', unit=' record(s)', leave=False,
                         disable=quiet, mininterval=0.2):
            num += 1
            bytes_written += b

        if not quiet:
            click.echo('Recompression completed.')
    except KeyboardInterrupt:
        if not quiet:
            click.echo('Recompression aborted.')
            sys.exit(1)
    finally:
        if not quiet and num > 0:
            click.echo(f'  - Records recompressed: {num}')
            click.echo(f'  - Bytes written: {_human_readable_bytes(bytes_written)}')
            click.echo(f'  - Completed in: {time.time() - start:.02f} seconds')


@main.command()
@click.argument('infile', type=click.Path(dir_okay=False, exists=True))
@click.option('-d', '--decompress-alg', type=click.Choice(['gzip', 'lz4', 'uncompressed', 'auto']),
              default='auto', show_default=True, help='Decompression algorithm')
@click.option('-p', '--verify-payloads', is_flag=True, help='Also verify payload digests')
@click.option('-q', '--quiet', is_flag=True, help='Do not print progress information')
@click.option('-o', '--output', type=click.File('w'), help='Output file with verification details')
def check(infile, decompress_alg, verify_payloads, quiet, output):
    """Verify WARC consistency by checking all digests."""

    decompress_alg = getattr(CompressionAlg, decompress_alg)

    failed_digests = []
    num_total = 0
    num_no_digest = 0
    try:
        desc_tpl = 'Verifying digests ({} failed)'
        pbar = tqdm(tools.verify_digests(infile, verify_payloads, decompress_alg),
                    desc=desc_tpl.format(0), unit=' record(s)', leave=False, disable=quiet, mininterval=0.2)

        for v in pbar:
            num_total += 1

            status = 'OK'
            if v['block_digest_ok'] is False:
                status = 'FAIL'
            elif v['block_digest_ok'] is None:
                status = 'NO_DIGEST'

            if 'payload_digest_ok' in v:
                if v['payload_digest_ok'] is True:
                    status += ', PAYLOAD_OK'
                elif v['payload_digest_ok'] is False:
                    status += ', PAYLOAD_FAIL'
                elif v['payload_digest_ok'] is None:
                    status += ', PAYLOAD_NO_DIGEST'

            if 'FAIL' in status:
                failed_digests.append(v['record_id'])
                pbar.set_description(desc_tpl.format(len(failed_digests)))
            elif 'NO_DIGEST' in status and 'OK' not in status:
                num_no_digest += 1

            if output:
                output.write(f'{v["record_id"]}: {status}\n')

        if output:
            output.close()

    except KeyboardInterrupt:
        if not quiet:
            click.echo('Verification aborted.')
            sys.exit(1)
    finally:
        if not quiet and num_total > 0:
            if not failed_digests:
                click.echo(f'{num_total - num_no_digest} records were verified successfully.')
                if num_no_digest:
                    click.echo(f'{num_no_digest} records were skipped without digest.')
            else:
                click.echo('Failed records:')
                click.echo('===============')
                for rec in failed_digests:
                    click.echo(rec)
                sys.exit(1)


main.add_command(benchmark)



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libc.stdio cimport FILE

from resiliparse_inc.zlib cimport *
from resiliparse_inc.lz4hc cimport *
from resiliparse_inc.lz4frame cimport *
from resiliparse_inc.string cimport string
from resiliparse_inc.string_view cimport string_view


cdef class IOStream:
    cdef string errstr

    cdef string read(self, size_t size)
    cdef size_t write(self, const char* data, size_t size)
    cdef size_t tell(self)
    cdef void flush(self)
    cdef void close(self)
    cdef string error(self)


# noinspection PyAttributeOutsideInit
cdef class PythonIOStreamAdapter(IOStream):
    """IOStream adapter for Python file-like objects."""

    cdef object py_stream
    cdef Exception exc

    cdef inline size_t tell(self):
        try:
            return self.py_stream.tell()
        except Exception as e:
            self.errstr = str(e).encode()
            self.exc = e
            return 0

    cdef inline string read(self, size_t size):
        try:
            return self.py_stream.read(size)[:size]
        except Exception as e:
            self.errstr = str(e).encode()
            self.exc = e
            return string()

    cdef inline size_t write(self, const char* data, size_t size):
        try:
            return self.py_stream.write(data[:size])
        except Exception as e:
            self.errstr = str(e).encode()
            self.exc = e
            return 0

    cdef inline void flush(self):
        try:
            self.py_stream.flush()
        except Exception as e:
            self.errstr = str(e).encode()
            self.exc = e

    cdef inline void close(self):
        if self.py_stream is None:
            return

        # noinspection PyBroadException
        try:
            self.py_stream.close()
        except:
            # Ignore exceptions during close()
            pass


cdef class BytesIOStream(IOStream):
    cdef string buffer
    cdef size_t pos

    cdef void seek(self, size_t offset)
    cdef string getvalue(self)


cdef class FileStream(IOStream):
    cdef FILE* fp

    cdef void open(self, const char* path, char* mode=*) except *
    cdef void seek(self, size_t offset)


cdef class CompressingStream(IOStream):
    cdef size_t begin_member(self)
    cdef size_t end_member(self)


cdef class GZipStream(CompressingStream):
    cdef IOStream raw_stream
    cdef string working_buf
    cdef z_stream zst
    cdef int stream_read_status
    cdef char initialized
    cdef size_t stream_pos
    cdef bint member_started
    cdef int compression_level

    cdef void prepopulate(self, bint deflate, const string& initial_data)
    cdef void _init_z_stream(self, bint deflate) nogil
    cdef void _free_z_stream(self) nogil
    cdef bint _refill_working_buf(self, size_t size) nogil


cdef class LZ4Stream(CompressingStream):
    cdef IOStream raw_stream
    cdef LZ4F_cctx* cctx
    cdef LZ4F_dctx* dctx
    cdef LZ4F_preferences_t prefs
    cdef string working_buf
    cdef bint frame_started
    cdef size_t stream_pos

    cdef void prepopulate(self, const string& initial_data)
    cdef void _free_ctx(self) nogil


cdef class BufferedReader:
    cdef IOStream stream
    cdef string errstr
    cdef string buf
    cdef size_t buf_size
    cdef size_t stream_pos
    cdef size_t limit
    cdef size_t limit_consumed
    cdef bint negotiate_stream
    cdef bint stream_started
    cdef bint stream_is_compressed

    cdef inline void set_limit(self, size_t offset) nogil
    cdef inline void reset_limit(self) nogil
    cdef void detect_stream_type(self)

    cpdef string read(self, size_t size=*)
    cpdef string readline(self, bint crlf=*, size_t max_line_len=*)
    cpdef size_t tell(self)
    cpdef void consume(self, size_t size=*)
    cpdef void close(self)
    cpdef string error(self)

    cdef bint _fill_buf(self)
    cdef inline string_view _get_buf(self) nogil
    cdef inline void _consume_buf(self, size_t size) nogil



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# distutils: language = c++

cimport cython
from cython.operator cimport dereference as deref, preincrement as inc
from libc.stdint cimport uint16_t
from libcpp.vector cimport vector

import base64
from datetime import datetime
import hashlib
from typing import Iterable, List, Optional, Tuple
import uuid
import warnings

from resiliparse_inc.cctype cimport isspace, tolower
from resiliparse_inc.cstdlib cimport strtol
from resiliparse_inc.string cimport npos as strnpos, string, to_string

from fastwarc.stream_io cimport BufferedReader, BytesIOStream, CompressingStream, IOStream, PythonIOStreamAdapter
from fastwarc.stream_io import StreamError


cdef string strip_str(const string& s) nogil:
    cdef size_t start = 0
    cdef size_t end = s.size()

    for start in range(0, s.size()):
        if not isspace(s[start]):
            break

    for end in reversed(range(s.size())):
        if not isspace(s[end]):
            break

    return s.substr(start, end - start + 1)


cdef string str_to_lower(string s) nogil:
    for i in range(s.size()):
        s[i] = tolower(s[i])
    return s


cdef const char* _enum_record_type_to_str(WarcRecordType record_type) nogil:
    if record_type == warcinfo:
        return b'warcinfo'
    elif record_type == response:
        return b'response'
    elif record_type == resource:
        return b'resource'
    elif record_type == request:
        return b'request'
    elif record_type == metadata:
        return b'metadata'
    elif record_type == revisit:
        return b'revisit'
    elif record_type == conversion:
        return b'conversion'
    elif record_type == continuation:
        return b'continuation'
    else:
        return b'unknown'


cdef WarcRecordType _str_record_type_to_enum(string record_type) nogil:
    record_type = str_to_lower(record_type)
    if record_type == b'warcinfo':
        return warcinfo
    elif record_type == b'response':
        return response
    elif record_type == b'resource':
        return resource
    elif record_type == b'request':
        return request
    elif record_type == b'metadata':
        return metadata
    elif record_type == b'revisit':
        return revisit
    elif record_type == b'conversion':
        return conversion
    elif record_type == b'continuation':
        return continuation
    else:
        return unknown


class CaseInsensitiveStr(str):
    """Case insensitive str implementation for use as dict key."""

    def __hash__(self):
        return hash(self.casefold())

    def __eq__(self, other):
        return self.casefold() == other.casefold()


class CaseInsensitiveStrDict(dict):
    """Case-insensitive str dict."""

    def __getitem__(self, key):
        return super().__getitem__(CaseInsensitiveStr(key))

    def __setitem__(self, key, value):
        super().__setitem__(CaseInsensitiveStr(key), value)

    def __contains__(self, key):
        return super().__contains__(CaseInsensitiveStr(key))

    def get(self, key, value=None):
        return super().get(CaseInsensitiveStr(key), value)

    def setdefault(self, key, value=None):
        return super().setdefault(CaseInsensitiveStr(key), value)

    def pop(self, key):
        return super().pop(CaseInsensitiveStr(key))

    def update(self, it=None, **kwargs):
        if hasattr(it, 'items'):
            it = it.items()
        if it is not None:
            for k, v in it:
                super().__setitem__(CaseInsensitiveStr(k), v)
        for k, v in kwargs.items():
            super().__setitem__(CaseInsensitiveStr(k), v)


# noinspection PyAttributeOutsideInit
cdef class WarcHeaderMap:
    """Dict-like type representing a WARC or HTTP header block"""

    def __init__(self, *args, **kwargs):
        """
        :param encoding: header source encoding
        :type encoding: str, optional, default: 'utf-8'
        """

    def __cinit__(self, encoding='utf-8'):
        self._enc = encoding
        self._dict_cache = None
        self._dict_cache_stale = True

    def __getitem__(self, header_key):
        return self.asdict()[header_key]

    def __setitem__(self, header_key, header_value):
        self.set_header(header_key.encode(self._enc, errors='ignore'),
                        header_value.encode(self._enc, errors='ignore'))

    def __iter__(self):
        yield from self.items()

    def __repr__(self):
        return str(self.asdict())

    def __contains__(self, item):
        return item in self.asdict()

    @property
    def status_line(self) -> str:
        """Header status line."""
        return self._status_line.decode(self._enc, errors='ignore')

    @status_line.setter
    def status_line(self, status_line):
        """Set status line contents."""
        self._status_line = status_line.encode(self._enc, errors='ignore')

    @property
    def status_code(self) -> int:
        """HTTP status code (unset if header block is not an HTTP header block)."""
        if self._status_line.find(<char*>b'HTTP/') != 0:
            return None
        s = self._status_line.split(b' ')
        if len(s) != 3 or not s[1].isdigit():
            return None
        return int(s)

    def append(self, key, value):
        """Append header (use if header name is not unique)."""
        self.append_header(key.encode(self._enc), value.encode(self._enc))

    def get(self, item, default=None) -> str:
        """Get header value or ``default``."""
        return self.asdict().get(item, default)

    def items(self):
        """Item view of keys and values."""
        return self.asdict().items()

    def keys(self):
        """Iterable of header keys."""
        return self.asdict().keys()

    def values(self):
        """Iterable of header values."""
        return self.asdict().values()

    def asdict(self) -> dict:
        """Headers as Python dict."""
        cdef str_pair h
        if self._dict_cache_stale:
            self._dict_cache = <dict>CaseInsensitiveStrDict({
                CaseInsensitiveStr(h[0].decode(self._enc, errors='ignore')): h[1].decode(self._enc, errors='ignore')
                for h in self._headers})
        return self._dict_cache

    def astuples(self) -> List[Tuple[str, str]]:
        """Headers as list of tuples (use if header keys are not necessarily unique)."""
        cdef str_pair h
        return [(h[0].decode(self._enc, errors='ignore'), h[1].decode(self._enc, errors='ignore'))
                for h in self._headers]


    cdef size_t write(self, IOStream stream):
        """Write header block into stream."""
        cdef size_t bytes_written = 0
        if not self._status_line.empty():
            bytes_written += stream.write(self._status_line.data(), self._status_line.size())
            bytes_written += stream.write(<char*>b'\r\n', 2)

        cdef vector[str_pair].iterator it = self._headers.begin()
        while it != self._headers.end():
            if not deref(it)[0].empty():
                bytes_written += stream.write(deref(it)[0].data(), deref(it)[0].size())
                bytes_written += stream.write(<char*>b': ', 2)
            bytes_written += stream.write(deref(it)[1].data(), deref(it)[1].size())
            bytes_written += stream.write(<char*>b'\r\n', 2)
            inc(it)
        return bytes_written

    cdef inline void clear(self):
        """Clear headers."""
        self._headers.clear()
        self._dict_cache_stale = True

    cdef inline void set_status_line(self, const string& status_line):
        """Set status line string."""
        self._status_line = status_line

    cdef string find_header(self, const string& header_key, const string& default):
        """Return value of first header occurrence or ``default`` (linear complexity)."""
        cdef string header_key_lower = str_to_lower(header_key)
        cdef vector[str_pair].iterator it = self._headers.begin()
        while it != self._headers.end():
            if str_to_lower(deref(it)[0]) == header_key_lower:
                return deref(it)[1]
            inc(it)

        return default

    cdef void set_header(self, const string& header_key, const string& header_value):
        """Set new header or overwrite existing header if it exists."""
        self._dict_cache_stale = True
        cdef string header_key_lower = str_to_lower(header_key)
        cdef vector[str_pair].iterator it = self._headers.begin()
        while it != self._headers.end():
            if str_to_lower(deref(it)[0]) == header_key_lower:
                deref(it)[1] = header_value
                return
            inc(it)

        self._headers.push_back((header_key, header_value))

    cdef inline void append_header(self, const string& header_key, const string& header_value):
        """Append header (use if header key is not unique)."""
        self._headers.push_back((header_key, header_value))
        self._dict_cache_stale = True

    cdef void add_continuation(self, const string& header_continuation_value):
        """Append value to previous header."""
        if not self._headers.empty():
            self._headers.back()[1].append(<char*>b' ')
            self._headers.back()[1].append(header_continuation_value)
        else:
            # This should no happen, but what can we do?!
            self._headers.push_back((<char*>b'', header_continuation_value))
        self._dict_cache_stale = True


# noinspection PyProtectedMember, PyAttributeOutsideInit
cdef class WarcRecord:
    """A WARC record."""

    def __cinit__(self):
        self._record_type = unknown
        self._is_http = False
        self._http_parsed = False
        self._content_length = 0
        self._headers = WarcHeaderMap.__new__(WarcHeaderMap, 'utf-8')
        self._http_headers = None
        self._stream_pos = 0

    @property
    def record_id(self) -> str:
        """Record ID (same as ``headers['WARC-Record'ID']``."""
        return self._headers['WARC-Record-ID']

    @property
    def record_type(self) -> WarcRecordType:
        """Record type (same as ``headers['WARC-Type']``."""
        return self._record_type

    @record_type.setter
    def record_type(self, WarcRecordType record_type):
        self._record_type = record_type
        self._headers['WARC-Type'] = _enum_record_type_to_str(record_type)

    @property
    def headers(self) -> WarcHeaderMap:
        """WARC record headers."""
        return self._headers

    @property
    def is_http(self) -> bool:
        """Whether record is an HTTP record"""
        return self._is_http

    @property
    def is_http_parsed(self) -> bool:
        """Whether HTTP headers have been parsed."""
        return self._http_parsed

    @property
    def http_headers(self) -> Optional[WarcHeaderMap]:
        """HTTP headers if record is an HTTP record and HTTP headers have been parsed yet."""
        return self._http_headers

    @property
    def http_charset(self) -> Optional[str]:
        """
        HTTP charset/encoding as returned by the server or ``None`` if no valid charset is set.
        A returned string is guaranteed to be a valid Python encoding name.
        """
        if not self._http_parsed or self._http_charset == <char*>b'_':
            return None

        if not self._http_charset.empty():
            return self._http_charset.decode()

        cdef string content_type = self._http_headers.find_header(<char*>b'content-type', <char*>b'')
        cdef size_t pos = content_type.find(<char*>b'charset=')
        if pos == strnpos:
            return None

        pos += 8
        cdef size_t pos_end = content_type.find(<char*>b';', pos)
        if pos_end != strnpos:
            pos_end = pos_end - pos

        self._http_charset = str_to_lower(strip_str(content_type.substr(pos, pos_end)))
        try:
            ''.encode(encoding=self._http_charset.decode())
        except LookupError:
            self._http_charset = <char*>b'_'
            return None

        return self._http_charset.decode()

    @property
    def content_length(self) -> int:
        """Remaining WARC length in bytes (not necessarily the same as the ``Content-Length`` header)."""
        return self._content_length

    @property
    def reader(self) -> BufferedReader:
        """Reader for the remaining WARC record content."""
        return self._reader

    @property
    def stream_pos(self) -> int:
        """WARC record start offset in the original (uncompressed) stream."""
        return self._stream_pos

    cpdef void init_headers(self, size_t content_length, WarcRecordType record_type=no_type, bytes record_urn=None):
        """
        Initialize mandatory headers in a fresh :class:`WarcRecord` instance.
        
        :param content_length: WARC record body length in bytes
        :type content_length: int
        :param record_type: WARC-Type
        :type record_type: WarcRecordType, optional, default: no_type
        :param record_urn: WARC-Record-ID as URN without ``'<'``, ``'>'`` (if unset, a random URN will be generated)
        :type record_urn: bytes, optional
        """
        if record_urn is None:
            record_urn = uuid.uuid4().urn.encode()

        if record_type == no_type:
            record_type = self.record_type
        if record_type == any_type or record_type == no_type:
            record_type = unknown
        self.record_type = record_type

        self._headers.clear()
        self._headers.set_status_line(<char*>b'WARC/1.1')
        self._headers.append_header(<char*>b'WARC-Type', _enum_record_type_to_str(record_type))
        self._headers.append_header(<char*>b'WARC-Date', datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ').encode())
        self._headers.append_header(<char*>b'WARC-Record-ID', b''.join((b'<', record_urn, b'>')))
        self._headers.append_header(<char*>b'Content-Length', to_string(content_length))

    cpdef void set_bytes_content(self, bytes b):
        """
        Set WARC body.
        
        :param b: body as bytes
        :type b: bytes
        """
        self._reader = BufferedReader.__new__(BufferedReader, BytesIOStream(b))
        self._content_length = len(b)

    cpdef void parse_http(self):
        """
        Parse HTTP headers and advance content reader.
        
        It is safe to call this method multiple times, even if the record is not an HTTP record.
        """
        if self._http_parsed or not self._is_http:
            return
        self._http_headers = WarcHeaderMap.__new__(WarcHeaderMap, 'iso-8859-15')
        cdef size_t num_bytes = parse_header_block(self.reader, self._http_headers, True)
        self._content_length = self._content_length - num_bytes
        self._http_parsed = True

    # noinspection PyTypeChecker
    cpdef size_t write(self, stream, bint checksum_data=False, size_t chunk_size=16384):
        """
        Write WARC record onto a stream.
        
        :param stream: output stream
        :param checksum_data: add block and payload digest headers
        :type checksum_data: bool, optional, default: False
        :param chunk_size: write block size
        :type chunk_size: int, optional, default: 16384
        :return: number of bytes written
        """
        # If the raw byte content hasn't been parsed, we can simply pass it through
        if not checksum_data and not self.http_parsed:
            return self._write_impl(self.reader, stream, True, chunk_size)

        # Otherwise read everything into memory for content-length correction and checksumming
        cdef BytesIOStream block_buf = BytesIOStream()

        block_digest = hashlib.sha1() if checksum_data else None
        payload_digest = hashlib.sha1() if checksum_data and self._http_parsed else None

        if self._http_parsed:
            self._http_headers.write(block_buf)
            block_buf.write(<char*>b'\r\n', 2)

            if checksum_data:
                block_digest.update(block_buf.getvalue())

        cdef string payload_data
        while True:
            payload_data = self.reader.read(chunk_size)
            if payload_data.empty():
                break

            if checksum_data:
                block_digest.update(payload_data.data()[:payload_data.size()])
                if payload_digest is not None:
                    payload_digest.update(payload_data.data()[:payload_data.size()])
            block_buf.write(payload_data.data(), payload_data.size())

        self._headers.set_header(<char*>b'Content-Length', to_string(block_buf.tell()))
        if checksum_data:
            if payload_digest is not None:
                self._headers.set_header(<char*>b'WARC-Payload-Digest', b'sha1:' + base64.b32encode(payload_digest.digest()))
            self._headers.set_header(<char*>b'WARC-Block-Digest', b'sha1:' + base64.b32encode(block_digest.digest()))

        block_buf.seek(0)
        return self._write_impl(block_buf, stream, False, chunk_size)

    cdef size_t _write_impl(self, in_reader, out_stream, bint write_payload_headers, size_t chunk_size):
        cdef IOStream out_stream_wrapped
        cdef size_t bytes_written = 0
        cdef bint compress_member_started = False

        if isinstance(out_stream, IOStream):
            out_stream_wrapped = <IOStream>out_stream

            if isinstance(out_stream, CompressingStream):
                bytes_written = (<CompressingStream>out_stream_wrapped).begin_member()
                compress_member_started = True

        elif isinstance(out_stream, object) and hasattr(out_stream, 'write'):
            out_stream_wrapped = PythonIOStreamAdapter.__new__(PythonIOStreamAdapter, out_stream)
        else:
            warnings.warn(f"Object of type '{type(out_stream).__name__}' is not a valid stream.", RuntimeWarning)
            return 0

        bytes_written += self._headers.write(out_stream_wrapped)
        bytes_written += out_stream_wrapped.write(b'\r\n', 2)

        if write_payload_headers and self._http_parsed:
            bytes_written += self._http_headers.write(out_stream_wrapped)
            bytes_written += out_stream_wrapped.write(b'\r\n', 2)

        cdef string data
        while True:
            data = in_reader.read(chunk_size)
            if data.empty():
                break
            bytes_written += out_stream_wrapped.write(data.data(), data.size())
        bytes_written += out_stream_wrapped.write(b'\r\n', 2)

        if compress_member_started:
            bytes_written += (<CompressingStream> out_stream_wrapped).end_member()

        return bytes_written

    cdef bint _verify_digest(self, const string& base32_digest, bint consume):
        cdef size_t sep_pos = base32_digest.find(b':')
        if sep_pos == strnpos:
            return False

        cdef string alg = base32_digest.substr(0, sep_pos)
        cdef bytes digest = base64.b32decode(base32_digest.substr(sep_pos + 1))

        if alg == b'sha1':
            h = hashlib.sha1()
        elif alg == b'md5':
            h = hashlib.md5()
        elif alg == b'sha256':
            h = hashlib.sha256()
        else:
            warnings.warn(f'Unsupported hash algorithm "{alg.decode()}".')
            return False

        cdef string block
        cdef BytesIOStream tee_stream
        cdef bint consume_override = consume

        if isinstance(self._reader.stream, BytesIOStream):
            # Stream is already a BytesIOStream, so we don't need to create another copy
            consume_override = True
            tee_stream = <BytesIOStream>self._reader.stream
        elif not consume:
            tee_stream = BytesIOStream()

        while True:
            block = self._reader.read(1024)
            if block.empty():
                break
            h.update(block)
            if not consume_override:
                tee_stream.write(block.data(), block.size())

        if not consume:
            tee_stream.seek(0)
            self._reader = BufferedReader.__new__(BufferedReader, tee_stream)

        return h.digest() == digest

    cpdef bint verify_block_digest(self, bint consume=False):
        """
        Verify whether record block digest is valid.
        
        :param consume: do not create an in-memory copy of the record stream
                        (will fully consume the rest of the record)
        :type consume: bool, optional, default: False
        :return: ``True`` if digest exists and is valid
        """
        return self._verify_digest(self._headers.find_header(<char*>b'WARC-Block-Digest', <char*>b''), consume)

    cpdef bint verify_payload_digest(self, bint consume=False):
        """
        Verify whether record payload digest is valid.
        
        :param consume: do not create an in-memory copy of the record stream
                        (will fully consume the rest of the record)
        :type consume: bool, optional, default: False
        :return: ``True`` if record is HTTP record and digest exists and is valid
        """
        if not self._http_parsed:
            return False
        return self._verify_digest(self._headers.find_header(<char*>b'WARC-Payload-Digest', <char*>b''), consume)


# noinspection PyProtectedMember
cdef size_t parse_header_block(BufferedReader reader, WarcHeaderMap target, bint has_status_line=False):
    """
    Helper function for parsing WARC or HTTP header blocks.
    
    :param reader: input reader
    :type reader: BufferedReader
    :param target: header map to fill
    :type reader: WarcHeaderMap
    :param has_status_line: whether first line is a status line or already a header
    :type has_status_line: bool, optional, default: False
    :return: number of bytes read from `reader`
    """
    cdef string line
    cdef string header_key, header_value
    cdef size_t delim_pos = 0
    cdef size_t bytes_consumed = 0

    while True:
        line = reader.readline()
        bytes_consumed += line.size()

        if line == b'\r\n' or line == b'':
            break

        if isspace(line[0]):
            # Continuation line
            target.add_continuation(strip_str(line))
            continue

        if has_status_line:
            target.set_status_line(strip_str(line))
            has_status_line = False
            continue

        delim_pos = line.find(b':')
        if delim_pos == strnpos:
            # Invalid header, try to preserve it as best we can
            target.add_continuation(strip_str(line))
        else:
            header_key = strip_str(line.substr(0, delim_pos))
            if delim_pos >= line.size():
                header_value = string()
            else:
                header_value = strip_str(line.substr(delim_pos + 1))
            target.append_header(header_key, header_value)

    return bytes_consumed


# noinspection PyProtectedMember, PyAttributeOutsideInit
@cython.auto_pickle(False)
cdef class ArchiveIterator:
    """WARC record stream iterator."""

    def __init__(self, *args, **kwargs):
        """
        Initialize WARC record iterator.

        :param stream: input stream (preferably an :class:`~fastwarc.stream_io.IOStream`,
                       but any file-like Python object is fine)
        :param parse_http: whether to parse HTTP records automatically (disable for better performance if not needed)
        :type parse_http: bool, optional, default: True
        :param record_types: bitmask of :class:`WarcRecordType` record types to return (others will be skipped)
        :type record_types: int, optional, default: any_type
        :param min_content_length: skip records with Content-Length less than this
        :type min_content_length: int, optional
        :param max_content_length: skip records with Content-Length large than this
        :type max_content_length: int, optional
        :param func_filter: Python callable taking a :class:`WarcRecord` and returning a ``bool``
                            for further record filtering
        :type func_filter: Callable, optional
        :param verify_digests: skip records which have no or an invalid block digest
        :type verify_digests: bool, optional, default: False
        """

    def __cinit__(self, stream, uint16_t record_types=any_type, bint parse_http=True,
                  size_t min_content_length=strnpos, size_t max_content_length=strnpos,
                  func_filter=None, bint verify_digests=False):
        self._set_stream(stream)
        self.record = None
        self.parse_http = parse_http
        self.verify_digests = verify_digests
        self.min_content_length = min_content_length
        self.max_content_length = max_content_length
        self.func_filter = func_filter
        self.record_type_filter = record_types

    def __iter__(self) -> Iterable[WarcRecord]:
        cdef _NextRecStatus status

        while True:
            status = self._read_next_record()
            if status == has_next:
                yield self.record
            elif status == eof:
                self.reader.close()
                return

    cdef _NextRecStatus _read_next_record(self) except _NextRecStatus.error:
        self.reader.detect_stream_type()

        if self.record is not None:
            self.reader.consume()
            self.reader.reset_limit()

        self.record = WarcRecord.__new__(WarcRecord)
        self.record._stream_pos = self.reader.tell()

        cdef string version_line
        while True:
            version_line = self.reader.readline()
            if not self.reader.error().empty():
                if isinstance(self.reader.stream, PythonIOStreamAdapter) and \
                        (<PythonIOStreamAdapter>self.reader.stream).exc is not None:
                    raise (<PythonIOStreamAdapter>self.reader.stream).exc
                raise StreamError(self.reader.error().decode())

            if version_line.empty():
                # EOF
                return eof

            if version_line == b'\r\n' or version_line == b'\n':
                # Consume empty lines
                if not self.reader.stream_is_compressed:
                    self.record._stream_pos += version_line.size()
                continue

            version_line = strip_str(version_line)
            if version_line == b'WARC/1.0' or version_line == b'WARC/1.1':
                # OK, continue with parsing headers
                self.record._headers.set_status_line(version_line)
                break
            else:
                # Not a WARC file or unsupported version
                return eof

        parse_header_block(self.reader, self.record._headers)

        cdef string hkey
        cdef size_t parse_count = 0
        cdef str_pair h
        cdef bint skip = False
        for h in self.record._headers._headers:
            hkey = str_to_lower(h[0])
            if hkey == b'warc-type':
                self.record._record_type = _str_record_type_to_enum(h[1])
                parse_count += 1
            elif hkey == b'content-length':
                self.record._content_length = strtol(h[1].c_str(), NULL, 10)
                parse_count += 1
            elif hkey == b'content-type' and h[1].find(b'application/http') == 0:
                self.record._is_http = True
                parse_count += 1

            if parse_count >= 3:
                break

        # Check if record is to be skipped
        skip |= (self.record._record_type & self.record_type_filter) == 0
        skip |= self.max_content_length != strnpos and self.record._content_length > self.max_content_length
        skip |= self.min_content_length != strnpos and self.record._content_length < self.max_content_length
        if not skip and self.func_filter is not None:
            # Execute expensive filters last and only if skip is not already true
            skip |= not self.func_filter(self.record)

        if skip:
            self.reader.consume(self.record._content_length)
            self.record = None
            return skip_next

        self.reader.set_limit(self.record._content_length)
        self.record._reader = self.reader

        if self.verify_digests and not self.record.verify_block_digest(False):
            self.reader.reset_limit()
            self.reader.consume(self.record._content_length)
            self.record = None
            return skip_next

        if self.parse_http and self.record._is_http:
            self.record.parse_http()

        return has_next

    cpdef void _set_stream(self, stream):
        """
        Replace underlying input stream.
        
        This method is for internal use and should not be called by external users.
        """
        if not isinstance(stream, IOStream):
            for attr in ('read', 'tell', 'close'):
                if not hasattr(stream, attr):
                    raise AttributeError(f"Object of type '{type(stream).__name__}' has no attribute '{attr}'.")
            stream = PythonIOStreamAdapter.__new__(PythonIOStreamAdapter, stream)

        cdef IOStream stream_ = <IOStream>stream
        self.reader = BufferedReader.__new__(BufferedReader, stream_)
        self.record = None


# noinspection PyProtectedMember
cpdef bint is_warc_10(WarcRecord record):
    """Filter predicate for checking if record is a WARC/1.0 record."""
    return record._headers._status_line == <char*>b'WARC/1.0'


# noinspection PyProtectedMember
cpdef bint is_warc_11(WarcRecord record):
    """Filter predicate for checking if record is a WARC/1.1 record."""
    return record._headers._status_line == <char*>b'WARC/1.1'


# noinspection PyProtectedMember
cpdef bint has_block_digest(WarcRecord record):
    """Filter predicate for checking if record has a block digest."""
    return not record._headers.find_header(<char*>b'WARC-Block-Digest', <char*>b'').empty()


# noinspection PyProtectedMember
cpdef bint has_payload_digest(WarcRecord record):
    """Filter predicate for checking if record has a payload digest."""
    return not record._headers.find_header(<char*>b'WARC-Payload-Digest', <char*>b'').empty()


# noinspection PyProtectedMember
cpdef bint is_http(WarcRecord record):
    """Filter predicate for checking if record is an HTTP record."""
    return record._is_http


# noinspection PyProtectedMember
cpdef bint is_concurrent(WarcRecord record):
    """Filter predicate for checking if record is concurrent to another record."""
    return not record._headers.find_header(<char*>b'WARC-Concurrent-To', <char*>b'').empty()



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# distutils: language = c++

cimport cython

import warnings

from resiliparse_inc.cstdlib cimport strtol
from resiliparse_inc.cstring cimport strerror
from resiliparse_inc.errno cimport errno
from resiliparse_inc.stdio cimport fclose, fflush, fopen, fread, fseek, ftell, fwrite, SEEK_SET
from resiliparse_inc.string cimport npos as strnpos, string


class FastWARCError(Exception):
    """Generic FastWARC exception."""


class StreamError(FastWARCError):
    """FastWARC stream error."""


@cython.auto_pickle(False)
cdef class IOStream:
    """IOStream base class."""

    cdef string read(self, size_t size):
        pass

    cdef size_t write(self, const char* data, size_t size):
        pass

    cdef size_t tell(self):
        pass

    cdef void flush(self):
        pass

    cdef void close(self):
        pass

    cdef string error(self):
        return self.errstr


# noinspection PyAttributeOutsideInit
@cython.auto_pickle(False)
cdef class BytesIOStream(IOStream):
    """IOStream that uses an in-memory buffer."""
    def __init__(self, *args, **kwargs):
        """
        :param initial_data: fill internal buffer with this initial data
        :type initial_data: bytes, optional
        """

    def __cinit__(self, bytes initial_data=b''):
        self.pos = 0
        self.buffer = initial_data

    cdef inline size_t tell(self):
        return self.pos

    cdef inline void seek(self, size_t offset):
        self.pos = min(self.buffer.size(), offset)

    cdef inline string read(self, size_t size):
        if self.pos >= self.buffer.size():
            return string()
        cdef string substr = self.buffer.substr(self.pos, size)
        self.seek(self.pos + size)
        return substr

    cdef inline size_t write(self, const char* data, size_t size):
        if self.pos + size > self.buffer.size():
            self.buffer.resize(self.pos + size)
        cdef size_t i
        for i in range(size):
            self.buffer[self.pos + i] = data[i]
        self.pos += size
        return size

    cdef inline void close(self):
        self.buffer.clear()

    cdef inline string getvalue(self):
        return self.buffer


# noinspection PyAttributeOutsideInit
@cython.auto_pickle(False)
cdef class FileStream(IOStream):
    """Fast alternative to Python file objects for local files."""

    def __init__(self, *args, **kwargs):
        """
        :param filename: input filename
        :type filename: str, optional
        :param mode: file open mode
        :type mode: str, optional, default: 'rb'
        """

    def __cinit__(self, str filename=None, str mode='rb'):
        self.fp = NULL
        if filename:
            self.open(filename.encode(), mode.encode())

    def __dealloc__(self):
        self.close()

    cdef void open(self, char* path, char* mode=b'rb') except *:
        if self.fp != NULL:
            self.close()

        self.fp = fopen(path, mode)
        if self.fp == NULL:
            self.errstr = strerror(errno)
            raise StreamError(self.errstr.decode())
        self.errstr.clear()

    cdef void seek(self, size_t offset):
        fseek(self.fp, offset, SEEK_SET)

    cdef size_t tell(self):
        return ftell(self.fp)

    cdef string read(self, size_t size):
        cdef string buf
        cdef size_t c
        with nogil:
            buf.resize(size)
            c = fread(buf.data(), sizeof(char), size, self.fp)
            if errno:
                self.errstr = strerror(errno)
                return string()
            buf.resize(c)
            return buf

    cdef size_t write(self, const char* data, size_t size):
        cdef size_t w = fwrite(data, sizeof(char), size, self.fp)
        if errno:
            self.errstr = strerror(errno)
            return 0
        return w

    cdef void flush(self):
        fflush(self.fp)

    cdef void close(self):
        if self.fp != NULL:
            fclose(self.fp)
            self.fp = NULL


@cython.auto_pickle(False)
cdef class PythonIOStreamAdapter(IOStream):
    def __init__(self, *args, **kwargs):
        """
        :param py_stream: input Python stream object
        """

    def __cinit__(self, py_stream):
        self.py_stream = py_stream
        self.exc = None


cpdef IOStream wrap_stream(raw_stream):
    """
    Wrap ``raw_stream`` into a :class:`PythonIOStreamAdapter` if it is a Python object or
    return ``raw_stream`` unmodified if it is a :class:`IOStream` already.
    
    :param raw_stream: stream to wrap
    :return: wrapped stream
    """
    if isinstance(raw_stream, IOStream):
        return raw_stream
    elif isinstance(raw_stream, object) and hasattr(raw_stream, 'read'):
        return PythonIOStreamAdapter.__new__(PythonIOStreamAdapter, raw_stream)
    else:
        warnings.warn(f"Object of type '{type(raw_stream).__name__}' is not a valid stream.", RuntimeWarning)
        return None


@cython.auto_pickle(False)
cdef class CompressingStream(IOStream):
    """Base class for compressed :class:`IOStream` types."""

    cdef size_t begin_member(self):
        """Begin compression member/frame (if not already started)."""
        return 0

    cdef size_t end_member(self):
        """End compression member/frame (if one has been started)."""
        return 0


cdef char _GZIP_DEFLATE = 1
cdef char _GZIP_INFLATE = 2


# noinspection PyAttributeOutsideInit
@cython.auto_pickle(False)
cdef class GZipStream(CompressingStream):
    """GZip :class:`IOStream` implementation."""

    def __init__(self, *args, **kwargs):
        """
        :param raw_stream: raw data stream
        :param compression_level: GZip compression level (for compression only)
        :type compression_level: int, optional, default: 9
        """

    def __cinit__(self, raw_stream, compression_level=Z_BEST_COMPRESSION):
        self.raw_stream = wrap_stream(raw_stream)
        self.member_started = False
        self.working_buf = string()
        self.initialized = 0
        self.stream_read_status = Z_STREAM_END
        self.stream_pos = self.raw_stream.tell()
        self.compression_level = compression_level

    def __dealloc__(self):
        self.close()

    cdef size_t tell(self):
        if self.initialized == _GZIP_INFLATE:
            return self.stream_pos
        return self.raw_stream.tell()

    cdef void _init_z_stream(self, bint deflate) nogil:
        """
        Reset internal state and initialize ``z_stream``.
        
        :param deflate: ``True`` for compression context, ``False`` for decompression context.
        """

        if self.initialized:
            return

        self.zst.opaque = Z_NULL
        self.zst.zalloc = Z_NULL
        self.zst.zfree = Z_NULL
        self.zst.next_in = NULL
        self.zst.next_out = NULL
        self.zst.avail_in = 0
        self.zst.avail_out = 0
        self.stream_read_status = Z_STREAM_END
        self.working_buf.clear()
        self.errstr.clear()

        if deflate:
            deflateInit2(&self.zst, self.compression_level, Z_DEFLATED, 16 + MAX_WBITS, 9, Z_DEFAULT_STRATEGY)
            self.initialized = _GZIP_DEFLATE
        else:
            inflateInit2(&self.zst, 16 + MAX_WBITS)
            self.initialized = _GZIP_INFLATE

    cdef void prepopulate(self, bint deflate, const string& initial_data):
        """
        Fill internal working buffer with initial data.
        Use if some initial data of the stream have already been consumed (e.g., for stream content negotiation).
        Has to be called before the first :meth:`read()`.
        
        :param deflate: ``True`` if ``data`` is uncompressed, ``False`` if ``data`` is compressed GZip data. 
        :param initial_data: data to prepopulate
        """
        self._init_z_stream(deflate)
        self.working_buf.append(initial_data)
        self.zst.next_in = <Bytef*>self.working_buf.data()
        self.zst.avail_in = self.working_buf.size()

    cdef void _free_z_stream(self) nogil:
        """Release internal state and reset working buffer."""
        if not self.initialized:
            return
        if self.initialized == _GZIP_DEFLATE:
            deflateEnd(&self.zst)
        elif self.initialized == _GZIP_INFLATE:
            inflateEnd(&self.zst)
        self.working_buf.clear()
        self.initialized = 0

    cdef bint _refill_working_buf(self, size_t size) nogil:
        self.working_buf.erase(0, self.working_buf.size())
        with gil:
            self.working_buf.append(self.raw_stream.read(max(1024u, size)))
            if self.working_buf.empty():
                # EOF
                self._free_z_stream()
                self.stream_pos = self.raw_stream.tell()
                return False

        self.zst.next_in = <Bytef*>self.working_buf.data()
        self.zst.avail_in = self.working_buf.size()
        return True

    cdef string read(self, size_t size):
        if self.initialized == _GZIP_DEFLATE:
            # Deflate in progress
            return string()

        if not self.initialized:
            self._init_z_stream(False)

        if self.zst.avail_in == 0 or self.working_buf.empty():
            if not self._refill_working_buf(size):
                return string()

        cdef string out_buf = string(size, <char>0)
        self.zst.next_out = <Bytef*>out_buf.data()
        self.zst.avail_out = out_buf.size()

        with nogil:
            self.stream_read_status = inflate(&self.zst, Z_NO_FLUSH)
            while self.zst.next_out == <Bytef*>out_buf.data() and (self.stream_read_status == Z_OK or
                                                                   self.stream_read_status == Z_BUF_ERROR):
                if self.zst.avail_in == 0:
                    if not self._refill_working_buf(size):
                        break

                self.stream_read_status = inflate(&self.zst, Z_NO_FLUSH)

        # Error
        if self.stream_read_status < 0 and self.stream_read_status != Z_BUF_ERROR:
            self._free_z_stream()
            self.errstr = <char*>b'Not a valid GZip stream'
            return string()

        if self.stream_read_status == Z_STREAM_END:
            # Member end
            self.stream_pos = self.raw_stream.tell() - self.working_buf.size() + \
                              (self.zst.next_in - <Bytef*>self.working_buf.data())
            inflateReset(&self.zst)

        if self.zst.avail_out > 0:
            out_buf.resize(self.zst.next_out - <Bytef*>out_buf.data())

        if out_buf.empty():
            # We may have hit a member boundary, try again
            return self.read(size)

        return out_buf

    cdef size_t write(self, const char* data, size_t size):
        if self.initialized == _GZIP_INFLATE:
            # Inflate in progress
            return 0

        if not self.initialized:
            self._init_z_stream(True)

        self.zst.next_in = <Bytef*>data
        self.zst.avail_in = size

        self.begin_member()
        cdef size_t written = 0
        cdef size_t bound = max(4096u, deflateBound(&self.zst, size))
        if self.working_buf.size() < bound or self.working_buf.size() / 8 > bound:
            self.working_buf.resize(bound)
        self.zst.next_out = <Bytef*>self.working_buf.data()
        self.zst.avail_out = self.working_buf.size()

        cdef int status = Z_OK
        cdef size_t written_so_far = self.zst.total_out
        with nogil:
            while self.zst.avail_in > 0 and status == Z_OK:
                status = deflate(&self.zst, Z_NO_FLUSH)
                if self.zst.avail_in > 0 and self.zst.avail_out == 0:
                    # Out buffer fully filled, but in buffer still holding data
                    self.working_buf.append(4096u, <char>0)
                    self.zst.next_out = <Bytef*>&self.working_buf.back() - 4096u
                    self.zst.avail_out = 4096u

        written += self.zst.total_out - written_so_far
        if written == 0:
            return 0

        return self.raw_stream.write(self.working_buf.data(), written)

    cdef size_t begin_member(self):
        self.member_started = True
        return 0

    cdef size_t end_member(self):
        if not self.member_started:
            return 0

        self.zst.avail_in = 0
        self.zst.next_in = NULL
        self.zst.next_out = <Bytef*>self.working_buf.data()
        self.zst.avail_out = self.working_buf.size()

        cdef size_t written_so_far = self.zst.total_out
        cdef int status = deflate(&self.zst, Z_FINISH)
        while status == Z_OK or status == Z_BUF_ERROR:
            if self.zst.avail_out == 0:
                # Need larger output buffer (unlikely to ever happen at this point)
                self.working_buf.resize(self.working_buf.size() + 4096u)
                self.zst.next_out = <Bytef*>&self.working_buf.back() - 4096u
                self.zst.avail_out = 4096u
            status = deflate(&self.zst, Z_FINISH)

        cdef size_t written = self.zst.total_out - written_so_far
        deflateReset(&self.zst)
        self.member_started = False

        if written == 0:
            return 0
        return self.raw_stream.write(self.working_buf.data(), written)

    cdef void flush(self):
        self.end_member()
        self.raw_stream.flush()

    cdef void close(self):
        self.end_member()
        self._free_z_stream()
        if self.raw_stream is not None:
            self.raw_stream.close()

    cdef string error(self):
        if not self.errstr.empty():
            return self.errstr
        elif self.raw_stream is not None:
            return self.raw_stream.error()
        return string()


# noinspection PyAttributeOutsideInit
@cython.auto_pickle(False)
cdef class LZ4Stream(CompressingStream):
    """LZ4 :class:`IOStream` implementation."""

    def __init__(self, *args, **kwargs):
        """
        :param raw_stream: raw data stream
        :param compression_level: LZ4 compression level (for compression only)
        :type compression_level: int, optional, default: 12
        :param favor_dec_speed: favor decompression speed over compression speed and size
        :type favor_dec_speed: bool, optional, default: True
        """

    def __cinit__(self, raw_stream, compression_level=LZ4HC_CLEVEL_MAX, favor_dec_speed=True):
        self.raw_stream = wrap_stream(raw_stream)
        self.cctx = NULL
        self.dctx = NULL
        self.working_buf = string()
        self.frame_started = False
        self.prefs.compressionLevel = compression_level
        self.prefs.favorDecSpeed = favor_dec_speed
        self.stream_pos = self.raw_stream.tell()

    def __dealloc__(self):
        self.close()

    cdef size_t tell(self):
        if self.dctx != NULL:
            return self.stream_pos
        return self.raw_stream.tell()

    cdef void prepopulate(self, const string& initial_data):
        """
        Fill internal working buffer with initial data.
        Use if some initial data of the stream have already been consumed (e.g., for stream content negotiation).
        Has to be called before the first :meth:`read()`.
        
        :param initial_data: data to prepopulate
        """
        self.working_buf.append(initial_data)

    cdef string read(self, size_t size):
        if self.cctx != NULL:
            # Decompression in progress
            return string()

        if self.working_buf.empty():
            self.working_buf.append(self.raw_stream.read(size))
            if self.working_buf.empty():
                # EOF
                self._free_ctx()
                return string()

        cdef string out_buf = string(size, <char>0)
        cdef size_t ret, bytes_read, bytes_written
        with nogil:
            bytes_read = self.working_buf.size()
            bytes_written = out_buf.size()

            if self.dctx == NULL:
                self.errstr.clear()
                LZ4F_createDecompressionContext(&self.dctx, LZ4F_VERSION)

            ret = LZ4F_decompress(self.dctx, out_buf.data(), &bytes_written,
                                  self.working_buf.data(), &bytes_read, NULL)
            while ret != 0 and bytes_written == 0 and not LZ4F_isError(ret):
                with gil:
                    self.working_buf.erase(0, bytes_read)
                    self.working_buf.append(self.raw_stream.read(size))
                if self.working_buf.empty():
                    # EOF
                    self._free_ctx()
                    return string()
                bytes_read = self.working_buf.size()
                bytes_written = out_buf.size()
                ret = LZ4F_decompress(self.dctx, out_buf.data(), &bytes_written,
                                      self.working_buf.data(), &bytes_read, NULL)

            if ret == 0:
                # Frame end
                with gil:
                    self.stream_pos = self.raw_stream.tell() - self.working_buf.size() + bytes_read
            elif LZ4F_isError(ret):
                self._free_ctx()
                self.errstr = <char*>b'Not a valid LZ4 stream'
                return string()
            self.working_buf.erase(0, bytes_read)

            if out_buf.size() != bytes_written:
                out_buf.resize(bytes_written)

        if out_buf.empty():
            # Everything OK, we may have hit a frame boundary
            return self.read(size)

        return out_buf

    cdef size_t begin_member(self):
        cdef size_t written
        with nogil:
            if self.cctx == NULL:
                self.errstr.clear()
                LZ4F_isError(LZ4F_createCompressionContext(&self.cctx, LZ4F_VERSION))

            if self.frame_started:
                return 0

            if self.working_buf.size() < LZ4F_HEADER_SIZE_MAX:
                self.working_buf.resize(LZ4F_HEADER_SIZE_MAX)
            written = LZ4F_compressBegin(self.cctx, self.working_buf.data(), self.working_buf.size(), &self.prefs)
            self.frame_started = True

        return self.raw_stream.write(self.working_buf.data(), written)

    cdef size_t end_member(self):
        cdef size_t written
        with nogil:
            if self.cctx == NULL or not self.frame_started:
                return 0
            written = LZ4F_compressEnd(self.cctx, self.working_buf.data(), self.working_buf.size(), NULL)
            self.frame_started = False
        return self.raw_stream.write(self.working_buf.data(), written)

    cdef size_t write(self, const char* data, size_t size):
        if self.dctx != NULL:
            # Compression in progress
            return 0

        cdef size_t buf_needed, written
        cdef size_t header_bytes_written = self.begin_member()
        with nogil:
            buf_needed = max(4096u, LZ4F_compressBound(size, &self.prefs))
            if self.working_buf.size() < buf_needed or self.working_buf.size() / 8 > buf_needed:
                self.working_buf.resize(buf_needed)

            written = LZ4F_compressUpdate(self.cctx, self.working_buf.data(), self.working_buf.size(),
                                          data, size, NULL)
        return self.raw_stream.write(self.working_buf.data(), written) + header_bytes_written

    cdef void flush(self):
        cdef size_t written
        cdef size_t buf_needed
        if self.cctx != NULL:
            buf_needed = LZ4F_compressBound(0, &self.prefs)
            if self.working_buf.size() < buf_needed:
                self.working_buf.resize(buf_needed)

            written = LZ4F_flush(self.cctx, self.working_buf.data(), self.working_buf.size(), NULL)
            self.raw_stream.write(self.working_buf.data(), written)
        self.raw_stream.flush()

    cdef void close(self):
        if self.cctx != NULL:
            self.end_member()

        self._free_ctx()
        if self.raw_stream is not None:
            self.raw_stream.close()

    cdef void _free_ctx(self) nogil:
        if self.cctx != NULL:
            LZ4F_freeCompressionContext(self.cctx)
            self.cctx = NULL

        if self.dctx != NULL:
            LZ4F_freeDecompressionContext(self.dctx)
            self.dctx = NULL

        if not self.working_buf.empty():
            self.working_buf.clear()

    cdef string error(self):
        if not self.errstr.empty():
            return self.errstr
        elif self.raw_stream is not None:
            return self.raw_stream.error()
        return string()


# noinspection PyAttributeOutsideInit
@cython.auto_pickle(False)
cdef class BufferedReader:
    """Buffered reader operating on an :class:`IOStream` instance."""

    def __init__(self, IOStream stream, size_t buf_size=16384, bint negotiate_stream=True):
        """
        Initialize :class:`BufferedReader`.

        :param stream: stream to operate on
        :type stream: IOStream
        :param buf_size: internal buffer size
        :type buf_size: int, optional, default: 16384
        :param negotiate_stream: whether to auto-negotiate stream type
        :type negotiate_stream: bool, optional, default: True
        """

    def __cinit__(self, IOStream stream, size_t buf_size=16384, bint negotiate_stream=True):
        self.stream = stream
        self.buf_size = max(1024u, buf_size)
        self.buf = string()
        self.limit = strnpos
        self.limit_consumed = 0
        self.stream_is_compressed = isinstance(stream, CompressingStream)
        self.stream_started = False
        self.negotiate_stream = negotiate_stream and not self.stream_is_compressed

    cdef void detect_stream_type(self):
        """
        Try to auto-detect stream type (GZip, LZ4, or uncompressed).
        """
        if not self.negotiate_stream or self.stream_started:
            return

        self.negotiate_stream = False
        if self.buf.empty():
            self._fill_buf()

        if self.buf.size() > 2 and self.buf[0] == <char> 0x1f and self.buf[1] == <char> 0x8b:
            self.stream = GZipStream.__new__(GZipStream, self.stream)
            (<GZipStream> self.stream).prepopulate(False, self.buf)
        elif self.buf.size() > 4 and self.buf.substr(0, 4) == <char*>b'\x04\x22\x4d\x18':
            self.stream = LZ4Stream.__new__(LZ4Stream, self.stream)
            (<LZ4Stream> self.stream).prepopulate(self.buf)
        elif self.buf.size() > 5 and self.buf.substr(0, 5) == <char*>b'WARC/':
            # Stream is uncompressed: bail out, dont' mess with buffers
            self.stream_is_compressed = False
            return
        else:
            self.errstr = <char*>b'Not a valid WARC stream'
            self.stream_is_compressed = False
            return

        self.stream_is_compressed = isinstance(self.stream, CompressingStream)
        self.buf.clear()
        self.buf.append(self.stream.read(self.buf_size))
        self.stream_started = False

    cdef bint _fill_buf(self):
        """
        Refill internal buffer.
        
        :return: ``True`` if refill was successful, ``False`` otherwise (EOF)
        """
        self.stream_started = True
        if self.buf.size() > 0:
            return True if self.limit == strnpos else self.limit > self.limit_consumed

        self.buf.append(self.stream.read(self.buf_size))
        return self.buf.size() > 0 if self.limit == strnpos else self.limit > self.limit_consumed

    cdef string_view _get_buf(self) nogil:
        """
        Get buffer contents. Does take a set limit into account.
        
        :return: available buffer contents
        """
        cdef string_view v = string_view(self.buf.c_str(), self.buf.size())
        cdef size_t remaining
        if self.limit != strnpos:
            remaining = self.limit - self.limit_consumed
            if v.size() > remaining:
                v.remove_suffix(v.size() - remaining)
        return v

    cdef void _consume_buf(self, size_t size) nogil:
        """
        Consume up to ``size`` bytes from internal buffer. Takes a set limit into account.
        
        :param size: number of bytes to read
        :type size: int
        """
        if self.limit == strnpos and size >= self.buf.size():
            self.buf.clear()
            return

        if self.limit != strnpos:
            if size > self.limit - self.limit_consumed:
                size = self.limit - self.limit_consumed
            self.limit_consumed += size

        self.buf.erase(0, size)

    cdef inline void set_limit(self, size_t offset) nogil:
        """
        Set a stream limit in bytes. Any read beyond this limit will act as if the stream reached EOF.
        A set limit can be reset by calling :meth:`reset_limit()`.
        
        :param offset: limit in bytes
        :type offset: int
        """
        self.limit = offset
        self.limit_consumed = 0

    cdef inline void reset_limit(self) nogil:
        """Reset any previously set stream limit."""
        self.limit = strnpos

    cpdef string read(self, size_t size=strnpos):
        """
        Read up to ``size`` bytes from the input stream.
        
        :param size: number of bytes to read (default means read remaining stream)
        :type size: int, optional, default: strnpos
        :return: consumed buffer contents as bytes (or empty string if EOF)
        """
        cdef string data_read
        cdef size_t missing = size
        cdef string_view buf_sub

        while (size == strnpos or data_read.size() < size) and self._fill_buf():
            missing = size - data_read.size()
            buf_sub = self._get_buf().substr(0, missing)
            data_read.append(<string>buf_sub)
            self._consume_buf(buf_sub.size())
        return data_read

    cpdef string readline(self, bint crlf=True, size_t max_line_len=4096):
        """
        Read a single line from the input stream.
        
        :param crlf: whether lines are separated by CRLF or LF
        :type crlf: bool, optional, default: True
        :param max_line_len: maximum line length (longer lines will still be consumed, but the return
                             value will not be larger than this)
        :type max_line_len: int, optional, default: 4096
        :return: line contents (or empty string if EOF)
        """
        cdef string line

        if not self._fill_buf():
            return string()

        cdef size_t capacity_remaining = max_line_len
        cdef string_view buf = self._get_buf()
        cdef char* newline_sep = b'\r\n' if crlf else '\n'
        cdef short newline_offset = 1 if crlf else 0
        cdef size_t pos = buf.find(newline_sep)
        cdef bint last_was_cr = False

        with nogil:
            while pos == strnpos:
                if capacity_remaining > 0:
                    line.append(<string>buf.substr(0, min(buf.size(), capacity_remaining)))
                    capacity_remaining -= line.size()

                # CRLF may be split
                last_was_cr = crlf and buf.back() == <char>b'\r'

                # Consume rest of line
                self._consume_buf(buf.size())

                with gil:
                    if not self._fill_buf():
                        break
                buf = self._get_buf()
                if last_was_cr and buf.front() == <char>b'\n':
                    if capacity_remaining:
                        line.append(<char*>b'\n')
                    self._consume_buf(1)
                    pos = strnpos
                    break

                pos = buf.find(newline_sep)

            if not buf.empty() and pos != strnpos:
                if capacity_remaining > 0:
                    line.append(<string>buf.substr(0, min(pos + 1 + newline_offset, capacity_remaining)))
                self._consume_buf(pos + 1 + newline_offset)

        return line

    cpdef size_t tell(self):
        """
        Offset on the input stream.
        
        :return: offset
        """
        if not self.stream_started:
            return 0

        if self.limit != strnpos:
            return self.limit_consumed

        if self.stream_is_compressed:
            # Compressed streams advance their position only on block boundaries.
            # The reader position inside the stream is meaningless.
            return self.stream.tell()

        return self.stream.tell() - self.buf.size()

    cpdef void consume(self, size_t size=strnpos):
        """
        Consume up to ``size`` bytes from the input stream without allocating a buffer for it.
        
        :param size: number of bytes to read (default means read remaining stream)
        :type size: int, optional, default: strnpos
        """
        cdef string_view buf
        cdef size_t bytes_to_consume

        while size > 0 and self._fill_buf():
            buf = self._get_buf()
            if buf.empty():
                break

            if size != strnpos:
                bytes_to_consume = min(buf.size(), size)
                self._consume_buf(bytes_to_consume)
                size -= bytes_to_consume
            else:
                self._consume_buf(buf.size())

    cpdef void close(self):
        """Close stream."""
        if self.stream is not None:
            self.stream.close()

    cpdef string error(self):
        """
        Return error message of last stream exception (or empty if no error has occurred).
        
        :return: error message or empty string
        """
        if not self.errstr.empty():
            return self.errstr
        if self.stream is not None:
            return self.stream.error()
        return string()


cpdef string read_http_chunk(BufferedReader reader):
    """
    Helper function for reading chunked HTTP payloads.
    
    Each call to this function will try to read the next chunk. In case of an error
    or EOF, an empty byte string will be returned.
    
    :param reader: input reader
    :type reader: BufferedReader
    :return: contents of the next chunk or empty string if EOF
    """
    cdef string header_line = reader.readline(True)
    cdef size_t chunk_size = strtol(header_line.substr(0, header_line.size() - 2).c_str(), NULL, 16)
    return reader.read(chunk_size + 2).substr(0, chunk_size)



#!/usr/bin/env python3
#
# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import chain
import getpass
import importlib
import os
import sys
import time
import urllib.request

import botocore.exceptions
import click
from tqdm import tqdm

from fastwarc.stream_io import FileStream, StreamError
from fastwarc.warc import ArchiveIterator, WarcRecordType
from fastwarc.tools import CompressionAlg, detect_compression_algorithm, wrap_warc_stream


@click.group()
def benchmark():
    """Benchmark FastWARC performance."""
    return 0


boto3 = None
botocore = None
s3 = None


def _init_s3(endpoint_url, aws_access_key,  aws_secret_key=None):
    global boto3, botocore, s3

    if s3 is not None:
        return

    try:
        boto3 = importlib.import_module('boto3')
        botocore = importlib.import_module('botocore')
    except ModuleNotFoundError:
        raise click.UsageError('Boto3 needs to be installed for S3 benchmarking.')

    if not aws_access_key:
        raise click.UsageError("s3:// URLs require '--aws-access-key' to be set.")
    if not aws_secret_key:
        aws_secret_key = getpass.getpass()

    s3 = boto3.resource('s3', endpoint_url=endpoint_url, aws_access_key_id=aws_access_key,
                        aws_secret_access_key=aws_secret_key)


# noinspection PyProtectedMember, HttpUrlsUsage
def _get_raw_stream_from_url(input_url, use_python_stream=False):
    if input_url.startswith('s3://'):
        assert s3 is not None
        try:
            s3_bucket, s3_object = input_url[5:].split('/', 1)
            stream = s3.Object(s3_bucket, s3_object).get()['Body']._raw_stream
        except botocore.exceptions.ClientError as e:
            click.echo(f'Error connecting to S3. Message: {e}.', err=True)
            sys.exit(1)

    elif input_url.startswith('http://') or input_url.startswith('https://'):
        stream = urllib.request.urlopen(input_url)

    else:
        if input_url.startswith('file://'):
            input_url = input_url[7:]
        if not os.path.isfile(input_url):
            click.echo(f"No such file or directory: '{input_url}", err=True)
            sys.exit(1)

        stream = FileStream(input_url, 'rb') if not use_python_stream else open(input_url, 'rb')

    return stream


def _expand_s3_prefixes(input_urls, endpoint_url, aws_access_key, aws_secret_key):
    urls_expanded = []
    for url in input_urls:
        if url.startswith('s3://'):
            _init_s3(endpoint_url, aws_access_key, aws_secret_key)
            s3_bucket, s3_object_prefix = url[5:].split('/', 1)
            urls_expanded.extend('/'.join(('s3:/', s3_bucket, o.key))
                                 for o in s3.Bucket(s3_bucket).objects.filter(Prefix=s3_object_prefix))
        else:
            urls_expanded.append(url)
    return urls_expanded


# noinspection PyPackageRequirements
@benchmark.command()
@click.argument('input_url', nargs=-1)
@click.option('-d', '--decompress-alg', type=click.Choice(['gzip', 'lz4', 'uncompressed', 'auto']),
              default='auto', show_default=True, help='Decompression algorithm')
@click.option('-e', '--endpoint-url', help='S3 endpoint URL', default='https://s3.amazonaws.com', show_default=True)
@click.option('-a', '--aws-access-key', help='AWS access key for s3:// URLs')
@click.option('-s', '--aws-secret-key', help='AWS secret key for s3:// URLs (leave empty to read from STDIN)')
@click.option('--is-prefix', is_flag=True, help='Treat input URL as prefix (only for S3)')
@click.option('-p', '--use-python-stream', is_flag=True,
              help='Use slower Python I/O instead of native FileStream for local files')
@click.option('-H', '--parse-http', is_flag=True, help='Parse HTTP headers')
@click.option('-v', '--verify-digests', is_flag=True, help='Verify record block digests')
@click.option('-f', '--filter-type', type=click.Choice(['warcinfo', 'response', 'resource', 'request', 'metadata',
                                                        'revisit', 'conversation', 'continuation', 'any_type']),
              default=['any_type'], multiple=True, show_default=True, help='Filter for specific WARC record types')
@click.option('-w', '--bench-warcio', is_flag=True, help='Compare FastWARC performance with WARCIO')
def read(input_url, decompress_alg, endpoint_url, aws_access_key, aws_secret_key, is_prefix, use_python_stream,
         verify_digests, parse_http, filter_type, bench_warcio):
    """
    Benchmark WARC read performance.

    Supported WARC sources are local files, S3 and HTTP(s) URLs.
    Supported compression algorithms are GZip, LZ4, or uncompressed.
    """

    if not input_url:
        raise click.UsageError('Need at least one input path.')

    if is_prefix:
        input_url = _expand_s3_prefixes(input_url, endpoint_url, aws_access_key, aws_secret_key)
    else:
        for u in input_url:
            if u.startswith('s3://'):
                _init_s3(endpoint_url, aws_access_key, aws_secret_key)
                break

    if decompress_alg == 'auto':
        decompress_alg = detect_compression_algorithm(input_url[0])
    else:
        decompress_alg = getattr(CompressionAlg, decompress_alg)

    if bench_warcio:
        if decompress_alg == CompressionAlg.lz4:
            click.echo('WARCIO does not support LZ4.', err=True)
            sys.exit(1)

        try:
            import warcio
        except ModuleNotFoundError:
            click.echo("--bench-warcio requires 'warcio' to be installed.", err=True)
            sys.exit(1)

    click.echo(f'Benchmarking read performance from {len(input_url)} input path(s)...')

    rec_type_filter = WarcRecordType.any_type
    if 'any_type' not in filter_type:
        rec_type_filter = WarcRecordType.no_type
        for t in filter_type:
            rec_type_filter |= getattr(WarcRecordType, t)

    def _bench(urls, stream_loader, lib):
        interrupted = False
        start = time.monotonic()
        num = 0

        try:
            def _load_lazy(u, l):
                yield from l(u)

            for _ in tqdm(chain(*(_load_lazy(u, stream_loader) for u in urls)),
                          desc=f'Benchmarking {lib}', unit=' records', leave=False, mininterval=0.3):
                num += 1
        except StreamError as e:
            click.echo(f'Error reading input: {e}', err=True)
            sys.exit(1)
        except KeyboardInterrupt:
            click.echo(f'Benchmark interrupted.', err=True)
            interrupted = True

        return num, time.monotonic() - start, interrupted

    def _fastwarc_iterator(f):
        s = _get_raw_stream_from_url(f, use_python_stream)
        s = wrap_warc_stream(s, 'rb', decompress_alg)
        return ArchiveIterator(s, rec_type_filter, parse_http=parse_http, verify_digests=verify_digests)

    n, t_fastwarc, interrupted = _bench(input_url, _fastwarc_iterator, 'FastWARC')
    click.echo(f'FastWARC: {n} records read in {t_fastwarc:.02f} seconds ({n / t_fastwarc:.02f} records/s).')

    if interrupted:
        sys.exit(1)

    if bench_warcio:
        def _warcio_iterator(f):
            s = _get_raw_stream_from_url(f, True)
            if rec_type_filter == WarcRecordType.any_type:
                yield from warcio.ArchiveIterator(s, no_record_parse=not parse_http, check_digests=verify_digests)
            else:
                # WARCIO does not support record filtering out of the box
                for rec in warcio.ArchiveIterator(s, not parse_http):
                    if rec_type_filter & getattr(WarcRecordType, rec.rec_type) != 0:
                        yield rec

        n, t_warcio, _ = _bench(input_url, _warcio_iterator, 'WARCIO')
        click.echo(f'WARCIO:   {n} records read in {t_warcio:.02f} seconds ({n / t_warcio:.02f} records/s).')
        click.echo(f'Time difference: {t_fastwarc - t_warcio:.02f} seconds, speedup: {t_warcio / t_fastwarc:.02f}')
